from typing import Any

from orjson import orjson


def dumps(content: Any, use_indent: bool = False, sort_keys: bool = False, **kwargs) -> bytes:
    option = orjson.OPT_SERIALIZE_UUID
    if use_indent:
        option |= orjson.OPT_INDENT_2
    if sort_keys:
        option |= orjson.OPT_SORT_KEYS
    return orjson.dumps(content, option=option, **kwargs)


def loads(data: bytes | bytearray | memoryview | str) -> Any:
    return orjson.loads(data)



import sys
from itertools import chain

import pydantic

from greyhorse.result import Result
from greyhorse.utils.injectors import ParamsInjector
from greyhorse.utils.invoke import invoke_sync
from greyhorse.utils.imports import import_path

from ..entities.application import Application
from ..entities.controller import Controller, ControllerFactoryFn
from ..entities.module import Module, ModuleErrorsItem
from ..entities.service import Service, ServiceFactoryFn
from ..errors import ControllerCreationError, CtrlFactoryNotFoundError, InvalidModuleConfError, ModuleCreationError, \
    ModuleLoadError, ModuleUnloadError, ModuleValidationError, ServiceCreationError, ServiceFactoryNotFoundError
from ..schemas.controller import ControllerConf
from ..schemas.module import ModuleConf, ModuleDesc
from ..schemas.service import ServiceConf
from ...i18n import tr
from ...logging import logger
import sys
from itertools import chain

import pydantic

from greyhorse.result import Result
from greyhorse.utils.imports import import_path
from greyhorse.utils.injectors import ParamsInjector
from greyhorse.utils.invoke import invoke_sync
from ..entities.application import Application
from ..entities.controller import Controller, ControllerFactoryFn
from ..entities.module import Module, ModuleErrorsItem
from ..entities.service import Service, ServiceFactoryFn
from ..errors import ControllerCreationError, CtrlFactoryNotFoundError, InvalidModuleConfError, ModuleCreationError, \
    ModuleLoadError, ModuleUnloadError, ModuleValidationError, ServiceCreationError, ServiceFactoryNotFoundError
from ..schemas.controller import ControllerConf
from ..schemas.module import ModuleConf, ModuleDesc
from ..schemas.service import ServiceConf
from ...i18n import tr
from ...logging import logger


class ModuleBuilder:
    def __init__(
        self, app: Application, root_desc: ModuleDesc,
    ):
        self._app = app
        self._injector = ParamsInjector()
        self._key_stack: list[str] = []
        self._root_desc = root_desc

    def get_module_init_path(self, desc: ModuleDesc) -> str:
        initpath = desc._initpath
        path = desc.path

        if desc.path.startswith('.'):
            dots_count = 0

            for c in path[1:]:
                if '.' == c:
                    dots_count += 1
                else:
                    path = path[dots_count + 1:]
                    break

            dots_count = min(dots_count, len(initpath))
            initpath = initpath[0:len(initpath) - dots_count]

        return '.'.join(initpath + [path]) if desc is not self._root_desc else desc.path

    def load_pass(self, desc: ModuleDesc | None = None) -> Result[ModuleConf | None]:
        desc = desc or self._root_desc

        if not desc.enabled:
            return Result.from_ok()

        res = self._load_module(desc)
        if not res.success:
            return res

        assert desc._conf
        conf = desc._conf

        if not conf.enabled:
            return Result.from_ok()

        for mod_desc in conf.submodules:
            res = self.load_pass(mod_desc)
            if not res.success:
                return res

        return Result.from_ok(conf)

    def create_module_pass(self, desc: ModuleDesc | None = None) -> Result[Module | None]:
        desc = desc or self._root_desc

        if not desc._conf:
            return Result.from_ok()

        current_conf: ModuleConf = desc._conf

        if not current_conf.enabled:
            logger.warn(tr('app.builder.module-disabled').format(path=desc.path))
            return Result.from_ok()

        logger.info(tr('app.builder.try-create').format(path=desc.path))

        values = {
            'name': current_conf.name,
            'conf': current_conf,
        }

        injected_args = self._injector(current_conf.factory, values=values)

        try:
            instance = invoke_sync(current_conf.factory, *injected_args.args, **injected_args.kwargs)
            assert isinstance(instance, Module)

        except Exception as e:
            error = ModuleCreationError(exc=e)
            logger.error(error.message)
            return Result.from_error(error)

        if errors := self._create_controllers(current_conf, instance):
            return Result.from_errors(list(chain.from_iterable([e.errors for e in errors])))

        if errors := self._create_services(current_conf, instance):
            return Result.from_errors(list(chain.from_iterable([e.errors for e in errors])))

        self._key_stack.append(current_conf.name)

        for submodule_desc in current_conf.submodules:
            res = self.create_module_pass(submodule_desc)
            if not res.success:
                return res
            if not res.result:
                continue

            submodule = res.result
            # noinspection PyProtectedMember
            submodule._resolve_claims(instance)
            instance.add_module(submodule.name, submodule)

        submodule_key = '.'.join(self._key_stack)
        self._app.register_module(submodule_key, instance, desc)
        self._key_stack.pop()

        logger.info(tr('app.builder.create-success').format(path=desc.path))
        return Result.from_ok(instance)

    def _load_module(self, desc: ModuleDesc) -> Result:
        if desc._conf is not None:
            return Result.from_ok()

        module_init_path = self.get_module_init_path(desc)
        func_name = module_init_path + ':__init__'
        logger.info(tr('app.builder.try-load').format(module_init_path=module_init_path))

        try:
            func = import_path(func_name)

        except (ImportError, AttributeError) as e:
            error = ModuleLoadError(exc=e)
            logger.error(error.message)
            return Result.from_error(error)

        injected_args = self._injector(func, values=desc.args)

        try:
            res = invoke_sync(func, *injected_args.args, **injected_args.kwargs)

        except pydantic.ValidationError as e:
            error = ModuleValidationError(detail=str(e))
            logger.error(error.message)
            return Result.from_error(error)
        except Exception as e:
            error = ModuleLoadError(exc=e)
            logger.error(error.message)
            return Result.from_error(error)

        if not isinstance(res, ModuleConf):
            error = InvalidModuleConfError(module_init_path=module_init_path)
            logger.error(error.message)
            return Result.from_error(error)

        logger.info(tr('app.builder.load-success').format(module_init_path=module_init_path))
        desc._conf = res
        return Result.from_ok()

    def _create_controller(
        self, module: Module, conf: ControllerConf, factory: ControllerFactoryFn,
    ) -> Result[Controller]:
        logger.info(tr('app.builder.try-ctrl').format(name=conf.name))
        values = {'name': conf.name, **conf.args}
        injected_args = self._injector(factory, values=values)

        try:
            res = factory(*injected_args.args, **injected_args.kwargs)

        except Exception as e:
            error = ControllerCreationError(exc=e)
            logger.error(error.message)
            return Result.from_error(error)

        if not res.success:
            return res

        instance = res.result

        res = instance.check_operator_mapping(conf.operator_mapping)
        if not res.success:
            return res

        key_mapping = res.result
        module.add_controller(
            conf.key, instance, name=conf.name,
            key_mapping=key_mapping,
            resources_read=conf.resources_read,
            resources_write=conf.resources_write,
            providers_read=conf.providers_read,
            operators_read=conf.operators_read,
        )

        logger.info(tr('app.builder.ctrl-success').format(name=conf.name))
        return Result.from_ok(instance)

    def _create_controllers(self, module_conf: ModuleConf, instance: Module) -> list[ModuleErrorsItem]:
        errors = []

        for conf in module_conf.controllers:
            if factory := module_conf.controller_factories.get(conf.key):
                res = self._create_controller(instance, conf, factory)
                if not res.success:
                    errors.append(ModuleErrorsItem(where='controller', name=conf.name, errors=res.errors))
            else:
                error = CtrlFactoryNotFoundError(key=str(conf.key.__name__))
                logger.error(error.message)
                errors.append(ModuleErrorsItem(where='controller', name=conf.name, errors=[error]))

        return errors

    def _create_service(
        self, module: Module, conf: ServiceConf, factory: ServiceFactoryFn,
    ) -> Result[Service]:
        logger.info(tr('app.builder.try-service').format(name=conf.name))
        values = {'name': conf.name, **conf.args}
        injected_args = self._injector(factory, values=values)

        try:
            res = factory(*injected_args.args, **injected_args.kwargs)

        except Exception as e:
            error = ServiceCreationError(exc=e)
            logger.error(error.message)
            return Result.from_error(error)

        if not res.success:
            return res

        instance = res.result

        res = instance.check_provider_mapping(conf.provider_mapping)
        if not res.success:
            return res

        prov_key_mapping = res.result

        res = instance.check_operator_mapping(conf.operator_mapping)
        if not res.success:
            return res

        op_key_mapping = res.result

        module.add_service(
            conf.key, instance, name=conf.name,
            op_key_mapping=op_key_mapping,
            prov_key_mapping=prov_key_mapping,
            resources_read=conf.resources_read,
            providers_read=conf.providers_read,
            operators_read=conf.operators_read,
        )

        submodule_key = '.'.join(self._key_stack)
        self._app.register_service(f'{submodule_key}.{conf.name}', instance)

        logger.info(tr('app.builder.service-success').format(name=conf.name))
        return Result.from_ok(instance)

    def _create_services(self, module_conf: ModuleConf, instance: Module) -> list[ModuleErrorsItem]:
        errors = []

        for conf in module_conf.services:
            if factory := module_conf.service_factories.get(conf.key):
                res = self._create_service(instance, conf, factory)
                if not res.success:
                    errors.append(ModuleErrorsItem(where='service', name=conf.name, errors=res.errors))
            else:
                error = ServiceFactoryNotFoundError(key=str(conf.key.__name__))
                logger.error(error.message)
                errors.append(ModuleErrorsItem(where='service', name=conf.name, errors=[error]))

        return errors


class ModuleTerminator:
    def __init__(
        self, app: Application, root_desc: ModuleDesc,
    ):
        self._app = app
        self._injector = ParamsInjector()
        self._key_stack: list[str] = []
        self._root_desc = root_desc

    def get_module_init_path(self, desc: ModuleDesc) -> str:
        return '.'.join(desc._initpath + [desc.path]) if desc is not self._root_desc else desc.path

    def unload_pass(self, desc: ModuleDesc | None = None) -> Result:
        desc = desc or self._root_desc

        if not desc.enabled or not desc._conf:
            return Result.from_ok()

        conf = desc._conf

        if not conf.enabled:
            return Result.from_ok()

        for mod_desc in conf.submodules:
            res = self.unload_pass(mod_desc)
            if not res.success:
                return res

        return self._unload_module(desc)

    def destroy_module_pass(
        self, desc: ModuleDesc | None = None, parent: Module | None = None,
    ) -> Result:
        desc = desc or self._root_desc

        if not desc._conf:
            return Result.from_ok()

        logger.info(tr('app.builder.try-destroy').format(path=desc.path))

        current_conf: ModuleConf = desc._conf
        self._key_stack.append(current_conf.name)
        submodule_key = '.'.join(self._key_stack)
        current = self._app.get_module(submodule_key)

        for submodule_desc in current_conf.submodules:
            res = self.destroy_module_pass(submodule_desc, current)
            if not res.success:
                return res

        if parent:
            parent.remove_module(current_conf.name)

        for srv in current.service_list:
            self._app.unregister_service(f'{submodule_key}.{srv.name}')

        self._app.unregister_module(submodule_key)
        self._key_stack.pop()

        logger.info(tr('app.builder.destroy-success').format(path=desc.path))
        return Result.from_ok()

    def _unload_module(self, desc: ModuleDesc) -> Result:
        if desc._conf is None:
            return Result.from_ok()

        module_init_path = self.get_module_init_path(desc)
        func_name = module_init_path + ':__fini__'
        logger.info(tr('app.builder.try-unload').format(module_init_path=module_init_path))

        try:
            func = import_path(func_name)

        except (ImportError, AttributeError):
            pass
        else:
            injected_args = self._injector(func, types={ModuleConf: desc._conf})

            try:
                invoke_sync(func, *injected_args.args, **injected_args.kwargs)

            except Exception as e:
                error = ModuleUnloadError(exc=e)
                logger.error(error.message)
                return Result.from_error(error)

        logger.info(tr('app.builder.unload-success').format(module_init_path=module_init_path))
        desc._conf = None
        del sys.modules[module_init_path]
        return Result.from_ok()



from typing import Pattern

from pydantic import BaseModel, ConfigDict, Field

from .components import ResourceConf
from .controller import OperatorMappingPolicy
from ..entities.providers import ProviderKey
from ..entities.service import ServiceKey


class ProviderMappingPolicy(BaseModel):
    model_config = ConfigDict(frozen=True)

    key: ProviderKey
    name_pattern: Pattern | None = None


class ServiceConf(ResourceConf[ServiceKey]):
    model_config = ConfigDict(frozen=True)

    operator_mapping: list[OperatorMappingPolicy] = Field(default_factory=list)
    provider_mapping: list[ProviderMappingPolicy] = Field(default_factory=list)



from typing import Any, Callable

from greyhorse.result import Result
from ..utils.registry import ReadonlyRegistry

type Operator = Any
type OperatorKey = type
type OperatorFactoryFn = Callable[[...], Result[Operator]]
type OperatorFactoryRegistry = ReadonlyRegistry[OperatorKey, OperatorFactoryFn]



from typing import Pattern

from pydantic import BaseModel, ConfigDict, Field

from .components import ResourceConf, ResourcePolicy
from ..entities.controller import ControllerKey
from ..entities.operator import OperatorKey


class OperatorMappingPolicy(BaseModel):
    model_config = ConfigDict(frozen=True)

    key: OperatorKey
    map_to: OperatorKey
    name_pattern: Pattern | None = None


class ControllerConf(ResourceConf[ControllerKey]):
    model_config = ConfigDict(frozen=True)

    operator_mapping: list[OperatorMappingPolicy] = Field(default_factory=list)
    resources_write: list[ResourcePolicy] = Field(default_factory=list)



from abc import ABC, abstractmethod
from contextlib import AbstractAsyncContextManager, AbstractContextManager, asynccontextmanager, contextmanager
from typing import Callable, Generic, TypeVar

SessionType = TypeVar('SessionType')
SyncSessionFactory = Callable[[], AbstractContextManager[SessionType]]
AsyncSessionFactory = Callable[[], AbstractAsyncContextManager[SessionType]]


class SyncEngine(Generic[SessionType], ABC):
    is_sync = True
    is_async = False

    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @contextmanager
    @abstractmethod
    def session(self, *args, **kwargs) -> AbstractContextManager[SessionType]:
        ...

    def teardown_session(self):
        pass

    @abstractmethod
    def start(self):
        ...

    @abstractmethod
    def stop(self):
        ...


class AsyncEngine(Generic[SessionType], ABC):
    is_sync = False
    is_async = True

    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @asynccontextmanager
    @abstractmethod
    async def session(self, *args, **kwargs) -> AbstractAsyncContextManager[SessionType]:
        ...

    async def teardown_session(self):
        pass

    @abstractmethod
    async def start(self):
        ...

    @abstractmethod
    async def stop(self):
        ...



from abc import ABC
from typing import Any, Mapping, Optional, Sequence, Set

from .model import Model
from ..serializers import Deserializer, Serializer


class SerializableModel(Model, ABC):
    class Meta(Model.Meta):
        serializer: Serializer = None
        deserializer: Deserializer = None

    def __init_subclass__(cls, **kwargs):
        from greyhorse.data.serializers.pickle import PickleSerializer, PickleDeserializer

        cls.Meta.serializer = cls.Meta.serializer or PickleSerializer()
        cls.Meta.deserializer = cls.Meta.deserializer or PickleDeserializer()

        super().__init_subclass__(**kwargs)

    # noinspection PyProtectedMember
    @classmethod
    def get_serializable_fields(cls) -> Set[str]:
        key = cls._fields_cache_key()
        if key not in cls.Meta._fields_cache:
            cls._calculate_fields()
        return cls.Meta._fields_cache[key].ser

    def get_serializable_values(self, only_fields: Sequence[str] = None) -> Optional[Mapping[str, Any]]:
        serializable_fields = self.get_serializable_fields()
        only_fields = set(only_fields) if only_fields else set()

        return {
            name: getattr(self, name) for name in serializable_fields
            if hasattr(self, name) and (not only_fields or name in only_fields)
        }

    def serialize(self, only_fields: Sequence[str] = None) -> bytes:
        return self.Meta.serializer.serialize(self.get_serializable_values(only_fields))

    @classmethod
    def deserialize(cls, data: bytes) -> Optional[Mapping[str, Any]]:
        return cls.Meta.deserializer.deserialize(data)



from dataclasses import dataclass
from typing import Any, Literal, Mapping, Self, TYPE_CHECKING, cast, override

from greyhorse.result import Error, Result
from greyhorse.utils.invoke import invoke_sync
from .controller import Controller, ControllerKey
from .deps import DepsOperator, DepsProvider
from .operator import OperatorFactoryFn, OperatorFactoryRegistry, OperatorKey
from .providers import ProviderFactoryFn, ProviderFactoryRegistry, ProviderKey
from .service import Service, ServiceKey
from ..errors import NoOpFoundForPattern, OpPolicyViolation, ProvClaimPolicyViolation
from ..schemas.components import OperatorPolicy, ProviderPolicy, ResourcePolicy
from ..utils.registry import DictRegistry, KeyMapping, ReadonlyRegistry
from ...i18n import tr
from ...logging import logger

if TYPE_CHECKING:
    from ..schemas.module import ModuleConf


@dataclass(slots=True, frozen=True)
class ModuleProviderItem:
    key: ProviderKey
    instance: ProviderFactoryFn
    name: str | None = None


@dataclass(slots=True, frozen=True)
class ModuleErrorsItem:
    where: Literal['module'] | Literal['controller'] | Literal['service'] | Literal['provider']
    name: str
    errors: list[Error] | None = None


class ModuleDepsProvider(DepsProvider):
    def __init__(
        self, op_registry: OperatorFactoryRegistry,
        prov_registry: ProviderFactoryRegistry,
        deps_registry: ReadonlyRegistry[Any, Any],
        resources_read: list[ResourcePolicy],
        providers_read: list[ProviderPolicy],
        operators_read: list[OperatorPolicy],
    ):
        self._op_registry = op_registry
        self._prov_registry = prov_registry
        self._deps_registry = deps_registry
        self._resources_read = {p.key: p.name_pattern for p in resources_read}
        self._providers_read = {p.key: p.name_pattern for p in providers_read}
        self._operators_read = {p.key: p.name_pattern for p in operators_read}

    @override
    def get_operator_factory(
        self, key: OperatorKey, name: str | None = None,
    ) -> OperatorFactoryFn | None:
        if key in self._operators_read:
            name_pattern = self._operators_read[key]
            if name_pattern is not None:
                if name is None or not name_pattern.match(name):
                    return None
            return self._op_registry.get(key, name)
        return None

    @override
    def get_provider_factory(
        self, key: ProviderKey, name: str | None = None,
    ) -> ProviderFactoryFn | None:
        if key in self._providers_read:
            name_pattern = self._providers_read[key]
            if name_pattern is not None:
                if name is None or not name_pattern.match(name):
                    return None
            return self._prov_registry.get(key, name)
        return None

    @override
    def get_resource(
        self, key: Any, name: str | None = None,
    ) -> Any | None:
        if key in self._resources_read:
            name_pattern = self._resources_read[key]
            if name_pattern is not None:
                if name is None or not name_pattern.match(name):
                    return None
            return self._deps_registry.get(key, name)
        return None


class ModuleDepsOperator(DepsOperator):
    def __init__(
        self, deps_registry: DictRegistry[Any, Any],
        resources_write: list[ResourcePolicy],
    ):
        self._deps_registry = deps_registry
        self._resources_write = {p.key: p.name_pattern for p in resources_write}

    @override
    def set_resource(
        self, key: Any, instance: Any, name: str | None = None,
    ) -> bool:
        if key in self._resources_write:
            name_pattern = self._resources_write[key]
            if name_pattern is not None:
                if name is None or not name_pattern.match(name):
                    return False
            return self._deps_registry.set(key, instance, name)
        return False


class Module:
    def __init__(self, name: str, conf: 'ModuleConf'):
        self._name = name
        self._conf = conf
        self._modules: dict[str, Self] = {}
        self._controllers = DictRegistry[ControllerKey, Controller]()
        self._services = DictRegistry[ServiceKey, Service]()
        self._provider_factories = DictRegistry[ProviderKey, ProviderFactoryFn]()
        self._operator_factories = DictRegistry[OperatorKey, OperatorFactoryFn]()
        self._deps_registry = DictRegistry[Any, Any]()

    @property
    def name(self) -> str:
        return self._name

    @property
    def config(self) -> 'ModuleConf':
        return self._conf

    def _create_module_deps_provider(
        self, resources_read: list[ResourcePolicy] | None = None,
        providers_read: list[ProviderPolicy] | None = None,
        operators_read: list[OperatorPolicy] | None = None,
    ):
        return ModuleDepsProvider(
            op_registry=self._operator_factories,
            prov_registry=self._provider_factories,
            deps_registry=self._deps_registry,
            resources_read=resources_read or [],
            providers_read=providers_read or [],
            operators_read=operators_read or [],
        )

    def add_controller(
        self, key: ControllerKey, instance: Controller, name: str | None = None,
        key_mapping: Mapping[OperatorKey, KeyMapping[OperatorKey]] | None = None,
        resources_read: list[ResourcePolicy] | None = None,
        resources_write: list[ResourcePolicy] | None = None,
        providers_read: list[ProviderPolicy] | None = None,
        operators_read: list[OperatorPolicy] | None = None,
    ):
        if self._controllers.set(key, instance, name=name):
            self._operator_factories.merge(instance.operator_factories, key_mapping)

        deps_prov = self._create_module_deps_provider(resources_read, providers_read, operators_read)
        instance.set_deps_provider(deps_prov)

        deps_op = ModuleDepsOperator(self._deps_registry, resources_write or [])
        instance.set_deps_operator(deps_op)

    def remove_controller(
        self, key: ControllerKey,
        key_mapping: Mapping[OperatorKey, KeyMapping[OperatorKey]] | None = None,
        name: str | None = None,
    ):
        if instance := self._controllers.get(key, name=name):
            self._operator_factories.subtract(instance.operator_factories, key_mapping)
        self._controllers.reset(key, name=name)
        instance.reset_deps_provider()

    def get_controller(self, key: ControllerKey, name: str | None = None) -> Controller | None:
        if instance := self._controllers.get(key, name=name):
            return cast(key, instance)
        return None

    def add_service(
        self, key: ServiceKey, instance: Service, name: str | None = None,
        op_key_mapping: Mapping[OperatorKey, KeyMapping[OperatorKey]] | None = None,
        prov_key_mapping: Mapping[ProviderKey, KeyMapping[ProviderKey]] | None = None,
        resources_read: list[ResourcePolicy] | None = None,
        providers_read: list[ProviderPolicy] | None = None,
        operators_read: list[OperatorPolicy] | None = None,
    ):
        if self._services.set(key, instance, name=name):
            self._operator_factories.merge(instance.operator_factories, op_key_mapping)
            self._provider_factories.merge(instance.provider_factories, prov_key_mapping)

        deps_prov = self._create_module_deps_provider(resources_read, providers_read, operators_read)
        instance.set_deps_provider(deps_prov)

    def remove_service(
        self, key: ServiceKey,
        op_key_mapping: Mapping[OperatorKey, KeyMapping[OperatorKey]] | None = None,
        prov_key_mapping: Mapping[ProviderKey, KeyMapping[ProviderKey]] | None = None,
        name: str | None = None,
    ):
        if instance := self._services.get(key, name=name):
            self._operator_factories.subtract(instance.operator_factories, op_key_mapping)
            self._provider_factories.subtract(instance.provider_factories, prov_key_mapping)
        self._services.reset(key, name=name)
        instance.reset_deps_provider()

    def get_service(self, key: ServiceKey, name: str | None = None) -> Service | None:
        if instance := self._services.get(key, name=name):
            return cast(key, instance)
        return None

    @property
    def module_names(self) -> list[str]:
        return list(self._modules.keys())

    @property
    def controllers_list(self) -> list[Controller]:
        result = []

        for key in self._controllers.list_keys():
            for name in self._controllers.get_names(key):
                result.append(self._controllers.get(key, name))

        return result

    @property
    def service_list(self) -> list[Service]:
        result = []

        for key in self._services.list_keys():
            for name in self._services.get_names(key):
                result.append(self._services.get(key, name))

        return result

    def add_module(self, name: str, instance: Self):
        if name not in self._modules:
            self._modules[name] = instance
            res = self._get_sub_exports(instance)
            if res.success:
                self._operator_factories.merge(instance.operator_factories, res.result)

    def remove_module(self, name: str):
        if instance := self._modules.pop(name, None):
            res = self._get_sub_exports(instance)
            if res.success:
                self._operator_factories.subtract(instance.operator_factories, res.result)

    def get_module(self, name) -> Self | None:
        return self._modules.get(name)

    @property
    def provider_factories(self) -> ProviderFactoryRegistry:
        return self._provider_factories

    @property
    def operator_factories(self) -> OperatorFactoryRegistry:
        return self._operator_factories

    def satisfy_provider_claims(self, items: list[ModuleProviderItem]) -> list[ModuleErrorsItem]:
        errors: list[ModuleErrorsItem] = []

        for item in items:
            self._provider_factories.set(item.key, item.instance, name=item.name)

        for module in self._modules.values():
            sub_errors = module._resolve_claims(self)
            for error_item in sub_errors:
                errors.append(ModuleErrorsItem(
                    where='module', name=module.name, errors=error_item.errors,
                ))

        return errors

    def _resolve_claims(self, parent: Self) -> list[ModuleErrorsItem]:
        errors: list[ModuleErrorsItem] = []

        for claim in self.config.provider_claims:
            result = False
            if claim.name_pattern is not None:
                for name in parent.provider_factories.get_names(claim.key):
                    if claim.name_pattern.match(name):
                        instance = parent.provider_factories.get(claim.key, name)
                        result = self._provider_factories.set(claim.key, instance, name=name)
                        break

            elif instance := parent.provider_factories.get(claim.key):
                result = self._provider_factories.set(claim.key, instance)

            if not result:
                claim_name = f'{claim.key}' + (f' ("{claim.name_pattern}")' if claim.name_pattern else '')
                error = ProvClaimPolicyViolation(type='module', name=self.name, claim_name=claim_name)
                errors.append(ModuleErrorsItem(
                    where='provider', name=claim_name, errors=[error],
                ))

        for module in self._modules.values():
            sub_errors = module._resolve_claims(self)
            for error_item in sub_errors:
                errors.append(ModuleErrorsItem(
                    where='module', name=module.name, errors=error_item.errors,
                ))

        return errors

    def _get_sub_exports(self, submodule: Self) -> Result[Mapping[OperatorKey, KeyMapping[OperatorKey]]]:
        policies_dict = {p.key: p for p in submodule.config.operator_exports}
        key_mapping = {}

        for key in submodule._operator_factories.list_keys():
            if key not in policies_dict:
                error = OpPolicyViolation(type='module', name=self.name, key=str(key.__name__))
                logger.error(error.message)
                return Result.from_error(error)

            name_pattern = policies_dict[key].name_pattern

            if name_pattern is None:
                key_mapping[key] = KeyMapping[OperatorKey](map_to=key)
                continue

            names = []

            for name in submodule._operator_factories.get_names(key):
                if name_pattern.match(name):
                    names.append(name)

            if not names:
                error = NoOpFoundForPattern(
                    type='module', name=self.name, key=str(key.__name__), pattern=name_pattern,
                )
                logger.error(error.message)
                return Result.from_error(error)

            key_mapping[key] = KeyMapping[OperatorKey](map_to=key, names=names)

        return Result.from_ok(key_mapping)

    def create(self) -> list[ModuleErrorsItem]:
        errors: list[ModuleErrorsItem] = []
        logger.info(tr('app.entities.module.create-start').format(name=self.name))

        for key in self._services.list_keys():
            for name in self._services.get_names(key):
                instance = self._services.get(key, name)
                res = invoke_sync(instance.create)
                if not res.success:
                    errors.append(ModuleErrorsItem(
                        where='service', name=name, errors=res.errors,
                    ))
                else:
                    logger.info(tr('app.entities.module.service-created').format(name=name))

        for module in self._modules.values():
            sub_errors = invoke_sync(module.create)
            for error_item in sub_errors:
                errors.append(ModuleErrorsItem(
                    where='module', name=module.name, errors=error_item.errors,
                ))

        for key in self._controllers.list_keys():
            for name in self._controllers.get_names(key):
                instance = self._controllers.get(key, name)
                res = invoke_sync(instance.create)
                if not res.success:
                    errors.append(ModuleErrorsItem(
                        where='controller', name=name, errors=res.errors,
                    ))
                else:
                    logger.info(tr('app.entities.module.ctrl-created').format(name=name))

        logger.info(tr('app.entities.module.create-finish').format(name=self.name))
        return errors

    def destroy(self) -> list[ModuleErrorsItem]:
        errors: list[ModuleErrorsItem] = []
        logger.info(tr('app.entities.module.destroy-start').format(name=self.name))

        for key in self._controllers.list_keys():
            for name in self._controllers.get_names(key):
                instance = self._controllers.get(key, name)
                res = invoke_sync(instance.destroy)
                if not res.success:
                    errors.append(ModuleErrorsItem(
                        where='controller', name=name, errors=res.errors,
                    ))
                else:
                    logger.info(tr('app.entities.module.ctrl-destroyed').format(name=name))

        for module in self._modules.values():
            sub_errors = invoke_sync(module.destroy)
            for error_item in sub_errors:
                errors.append(ModuleErrorsItem(
                    where='module', name=module.name, errors=error_item.errors,
                ))

        for key in self._services.list_keys():
            for name in self._services.get_names(key):
                instance = self._services.get(key, name)
                res = invoke_sync(instance.destroy)
                if not res.success:
                    errors.append(ModuleErrorsItem(
                        where='service', name=name, errors=res.errors,
                    ))
                else:
                    logger.info(tr('app.entities.module.service-destroyed').format(name=name))

        logger.info(tr('app.entities.module.destroy-finish').format(name=self.name))
        return errors

    def start(self):
        logger.info(tr('app.entities.module.start-start').format(name=self.name))

        for key in self._services.list_keys():
            for name in self._services.get_names(key):
                instance = self._services.get(key, name)
                invoke_sync(instance.start)
                logger.info(tr('app.entities.module.service-started').format(name=name))

        for module in self._modules.values():
            module.start()

        for key in self._controllers.list_keys():
            for name in self._controllers.get_names(key):
                instance = self._controllers.get(key, name)
                invoke_sync(instance.start)
                logger.info(tr('app.entities.module.ctrl-started').format(name=name))

        logger.info(tr('app.entities.module.start-finish').format(name=self.name))

    def stop(self):
        logger.info(tr('app.entities.module.stop-start').format(name=self.name))

        for key in self._controllers.list_keys():
            for name in self._controllers.get_names(key):
                instance = self._controllers.get(key, name)
                invoke_sync(instance.stop)
                logger.info(tr('app.entities.module.ctrl-stopped').format(name=name))

        for module in reversed(self._modules.values()):
            module.stop()

        for key in self._services.list_keys():
            for name in self._services.get_names(key):
                instance = self._services.get(key, name)
                invoke_sync(instance.stop)
                logger.info(tr('app.entities.module.service-stopped').format(name=name))

        logger.info(tr('app.entities.module.stop-finish').format(name=self.name))



from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Mapping, Self, override


@dataclass
class KeyMapping[K]:
    map_to: K
    names: list[str] = field(default_factory=list)


class ReadonlyRegistry[K, V](ABC):
    @abstractmethod
    def get_names(self, key: K) -> list[str]:
        ...

    @abstractmethod
    def has(self, key: K, name: str | None = None) -> bool:
        ...

    @abstractmethod
    def get(self, key: K, name: str | None = None) -> V | None:
        ...


class Registry[K, V](ReadonlyRegistry[K, V], ABC):
    @abstractmethod
    def set(self, key: K, instance: V, name: str | None = None) -> bool:
        ...

    @abstractmethod
    def reset(self, key: K, name: str | None = None) -> bool:
        ...

    @abstractmethod
    def merge(self, other: Self, key_mapping: Mapping[K, KeyMapping[K]] | None = None):
        ...

    @abstractmethod
    def subtract(self, other: Self, key_mapping: Mapping[K, KeyMapping[K]] | None = None):
        ...


class DictRegistry[K, V](Registry[K, V]):
    def __init__(self):
        self._registry: dict[K, dict[str, V]] = defaultdict(dict)

    def list_keys(self) -> list[K]:
        return list(self._registry.keys())

    @override
    def get_names(self, key: K) -> list[str]:
        if key in self._registry:
            return list(self._registry[key].keys())
        else:
            return []

    @override
    def has(self, key: K, name: str | None = None) -> bool:
        if key in self._registry:
            if name is None:
                return True
            else:
                return name in self._registry[key]
        else:
            return False

    @override
    def get(self, key: K, name: str | None = None) -> V | None:
        if key in self._registry:
            if name is None:
                values = list(self._registry[key].values())
                return values[0] if len(values) > 0 else None
            else:
                return self._registry[key].get(name)
        else:
            return None

    @override
    def set(self, key: K, instance: V, name: str | None = None) -> bool:
        if key in self._registry:
            if name is None:
                return instance is self._registry[key]['']
            elif name in self._registry[key]:
                return instance is self._registry[key][name]
        self._registry[key][name or ''] = instance
        return True

    @override
    def reset(self, key: K, name: str | None = None) -> bool:
        if key in self._registry:
            if name is None:
                del self._registry[key]
                return True
            elif name in self._registry[key]:
                if len(self._registry[key]) == 1:
                    del self._registry[key]
                else:
                    del self._registry[key][name]
                return True
        return False

    @override
    def merge(self, other: Self, key_mapping: Mapping[K, KeyMapping[K]] | None = None):
        for key, values_dict in other._registry.items():
            only_names = set()
            if key_mapping is not None:
                if key not in key_mapping:
                    continue
                else:
                    only_names = set(key_mapping[key].names)
                    key = key_mapping[key].map_to

            for name, value in values_dict.items():
                if not only_names or name in only_names:
                    self.set(key, value, name=name)

    @override
    def subtract(self, other: Self, key_mapping: Mapping[K, KeyMapping[K]] | None = None):
        for key, values_dict in other._registry.items():
            only_names = set()
            if key_mapping is not None:
                if key not in key_mapping:
                    continue
                else:
                    only_names = set(key_mapping[key].names)
                    key = key_mapping[key].map_to

            for name in values_dict.keys():
                if not only_names or name in only_names:
                    self.reset(key, name=name)



from ..result import Error, ErrorKwargsMixin


class GreyhorseError(ErrorKwargsMixin, Error):
    app = 'greyhorse'


class ModuleLoadError(GreyhorseError):
    type = 'app.builder.load-error'


class ModuleUnloadError(GreyhorseError):
    type = 'app.builder.unload-error'


class ModuleValidationError(GreyhorseError):
    type = 'app.builder.validation-error'


class InvalidModuleConfError(GreyhorseError):
    type = 'app.builder.invalid-module-conf'


class ControllerCreationError(GreyhorseError):
    type = 'app.builder.ctrl-error'


class CtrlFactoryNotFoundError(GreyhorseError):
    type = 'app.builder.ctrl-not-found'


class ServiceCreationError(GreyhorseError):
    type = 'app.builder.service-error'


class ServiceFactoryNotFoundError(GreyhorseError):
    type = 'app.builder.service-not-found'


class ModuleCreationError(GreyhorseError):
    type = 'app.builder.module-error'


class AppNotLoadedError(RuntimeError):
    pass


class OpPolicyViolation(GreyhorseError):
    type = 'app.entities.operator-policy-violation'


class NoOpFoundForPattern(GreyhorseError):
    type = 'app.entities.operator-pattern-violation'


class ProvPolicyViolation(GreyhorseError):
    type = 'app.entities.provider-policy-violation'


class NoProvFoundForPattern(GreyhorseError):
    type = 'app.entities.provider-pattern-violation'


class ProvClaimPolicyViolation(GreyhorseError):
    type = 'app.entities.prov-claim-policy-violation'



import asyncio
import contextlib
import threading
from contextvars import ContextVar
from typing import Any, Mapping


class ContextData(object):
    def __init__(self, props: Mapping[str, Any] | None = None, **kwargs):
        if props:
            for name, value in props.items():
                super().__setattr__(name, self._wrap(value))
        super().__setattr__('data', kwargs)
        self._id = None

    def _wrap(self, value):
        if isinstance(value, (tuple, list, set, frozenset)):
            return type(value)([self._wrap(v) for v in value])
        else:
            return value

    def __setattr__(self, key, value):
        super().__setattr__(key, self._wrap(value))

    def __contains__(self, item):
        return item in self.__dict__

    @property
    def raw_id(self):
        return self._id

    @property
    def session_id(self):
        if self._id:
            return self._id
        try:
            return id(asyncio.current_task())
        except RuntimeError:
            return threading.current_thread().ident

    @session_id.setter
    def session_id(self, value):
        self._id = value


_ctx: ContextVar[ContextData] = ContextVar('_ctx')


def get_context():
    if ctx := _ctx.get(None):
        return ctx
    _ctx.set(ContextData())
    return _ctx.get()


@contextlib.contextmanager
def with_context(force_new: bool = False):
    if not force_new:
        if ctx := _ctx.get(None):
            yield ctx
            return

    token = _ctx.set(ContextData())
    yield _ctx.get(None)
    _ctx.reset(token)



import asyncio
import signal
import threading
from asyncio import get_running_loop
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Collection, Optional, TYPE_CHECKING

from greyhorse.result import Result
from .controller import Controller, ControllerKey
from .module import Module, ModuleErrorsItem, ModuleProviderItem
from .service import Service, ServiceKey
from ..errors import AppNotLoadedError
from ...i18n import tr
from ...logging import logger
from ...utils.invoke import get_asyncio_loop

if TYPE_CHECKING:
    from ..schemas.module import ModuleDesc


class Application:
    def __init__(
        self, name: str, version: str = '', debug: bool = False,
    ):
        self._name = name
        self._version = version
        self._debug = debug
        self._path = self._inspect_cwd()
        self._modules: dict[str, tuple[Module, 'ModuleDesc']] = {}
        self._services: dict[str, Service] = {}
        self._root_desc: Optional['ModuleDesc'] = None
        self._module: Module | None = None

    @staticmethod
    def _inspect_cwd():
        import inspect

        for frame in reversed(inspect.stack()):
            path = Path(frame.filename).absolute()

            while path.parent != path:
                path = path.parent
                pyproject_toml_path = path / 'pyproject.toml'
                if pyproject_toml_path.exists():
                    return path

    @property
    def name(self) -> str:
        return self._name

    @property
    def version(self) -> str:
        return self._version

    @property
    def debug(self) -> bool:
        return self._debug

    def get_cwd(self) -> Path:
        return self._path.absolute()

    def register_module(self, name: str, instance: Module, desc: 'ModuleDesc'):
        self._modules[name] = (instance, desc)

    def unregister_module(self, name: str):
        self._modules.pop(name, None)

    def register_service(self, name: str, instance: Service):
        self._services[name] = instance

    def unregister_service(self, name: str):
        self._services.pop(name, None)

    def get_module(self, name: str) -> Module | None:
        if value := self._modules.get(name):
            return value[0]
        return None

    def get_controller(self, key: ControllerKey, name: str | None = None) -> Controller | None:
        if self._module:
            return self._module.get_controller(key, name=name)
        return None

    def get_service(self, key: ServiceKey, name: str | None = None) -> Service | None:
        if self._module:
            return self._module.get_service(key, name=name)
        return None

    def satisfy_provider_claims(self, items: list[ModuleProviderItem]) -> list[ModuleErrorsItem]:
        if self._module:
            return self._module.satisfy_provider_claims(items)
        raise AppNotLoadedError()

    def create(self) -> list[ModuleErrorsItem]:
        if self._module:
            return self._module.create()
        raise AppNotLoadedError()

    def destroy(self) -> list[ModuleErrorsItem]:
        if self._module:
            return self._module.destroy()
        raise AppNotLoadedError()

    def start(self):
        if self._module:
            return self._module.start()
        raise AppNotLoadedError()

    def stop(self):
        if self._module:
            return self._module.stop()
        raise AppNotLoadedError()

    def load(self, module_path: str, args: dict[str, Any] | None = None) -> Result:
        from ..builders.module import ModuleBuilder
        from ..schemas.module import ModuleDesc

        logger.info(tr('app.application.try-load').format(module_path=module_path))

        root_desc = ModuleDesc(path=module_path, args=args or {}, _initpath='')
        builder = ModuleBuilder(self, root_desc)

        res = builder.load_pass()
        if not res.success:
            return res

        self._root_desc = root_desc

        res = builder.create_module_pass()
        if not res.success:
            return res

        self._module = res.result

        logger.info(tr('app.application.load-success').format(module_path=module_path))
        return Result.from_ok()

    def unload(self) -> Result:
        from ..builders.module import ModuleTerminator

        logger.info(tr('app.application.try-unload').format(module_path=self._root_desc.path))

        terminator = ModuleTerminator(self, self._root_desc)

        res = terminator.destroy_module_pass()
        if not res.success:
            return res

        self._module = None

        res = terminator.unload_pass()
        if not res.success:
            return res

        logger.info(tr('app.application.unload-success').format(module_path=self._root_desc.path))
        self._root_desc = None
        return Result.from_ok()

    def run_sync(self, callback: Callable[[], None] | None = None):
        sync_events: list[threading.Event] = []
        async_events: list[asyncio.Event] = []

        for name, srv in self._services.items():
            match srv.wait():
                case threading.Event() as event:
                    sync_events.append(event)
                case asyncio.Event() as event:
                    async_events.append(event)

        all_events = sync_events + async_events

        async def waiter():
            async with asyncio.TaskGroup() as tg:
                for e in async_events:
                    tg.create_task(e.wait())

        logger.info(tr('app.application.run-sync-start').format(name=self.name))

        while not all([e.is_set() for e in all_events]):
            if async_events:
                loop = get_running_loop()
                loop.run_until_complete(waiter())

            sync_events_bools = [e.wait(0.1) for e in sync_events]

            if callback and not all(sync_events_bools):
                callback()

        logger.info(tr('app.application.run-sync-stop').format(name=self.name))

    async def run_async(self):
        sync_events: list[threading.Event] = []
        async_events: list[asyncio.Event] = []

        for name, srv in self._services.items():
            match srv.wait():
                case threading.Event() as event:
                    sync_events.append(event)
                case asyncio.Event() as event:
                    async_events.append(event)

        all_events = sync_events + async_events

        logger.info(tr('app.application.run-async-start').format(name=self.name))

        while not all([e.is_set() for e in all_events]):
            async with asyncio.TaskGroup() as tg:
                for e in sync_events:
                    tg.create_task(asyncio.to_thread(e.wait, 0.1))
                for e in async_events:
                    tg.create_task(e.wait())

        logger.info(tr('app.application.run-async-stop').format(name=self.name))

    @contextmanager
    def graceful_exit(self, signals: Collection[int] = (signal.SIGINT, signal.SIGTERM)):
        signals = set(signals)
        flag: list[bool] = []

        if loop := get_asyncio_loop():
            for sig_num in signals:
                loop.add_signal_handler(sig_num, self._exit_handler, sig_num, flag)
            try:
                yield
            finally:
                for sig_num in signals:
                    loop.remove_signal_handler(sig_num)
        else:
            original_handlers = []

            for sig_num in signals:
                original_handlers.append(signal.getsignal(sig_num))
                signal.signal(sig_num, self._exit_handler)
            try:
                yield
            finally:
                for sig_num, handler in zip(signals, original_handlers):
                    signal.signal(sig_num, handler)

    def _exit_handler(self, sig_num: 'signal.Signals', flag: list[bool], *_) -> None:
        if flag:
            self._second_exit_stage(sig_num)
        else:
            self._first_exit_stage(sig_num)
            flag.append(True)

    def _first_exit_stage(self, sig_num: 'signal.Signals'):
        fail = False

        try:
            self.stop()
        except RuntimeError:
            fail = True

        if fail:
            self._second_exit_stage(sig_num)

    def _second_exit_stage(self, sig_num: 'signal.Signals'):
        raise SystemExit(128 + sig_num)



from abc import ABC, abstractmethod
from typing import Awaitable, Callable


class Resource(ABC):
    @property
    @abstractmethod
    def active(self) -> bool:
        ...

    @abstractmethod
    def usages(self) -> int:
        ...

    @abstractmethod
    def start(self) -> Awaitable[None] | None:
        ...

    @abstractmethod
    def stop(self) -> Awaitable[None] | None:
        ...


type ResourceStartFn[T] = Callable[[T], Awaitable[None] | None]
type ResourceStopFn[T] = Callable[[T], Awaitable[None] | None]
type ResourceActivityFn[T] = Callable[[T], bool]
type ResourceUsageFn[T] = Callable[[T], int]


class ResourceAdapter[T](Resource):
    def __init__(
        self, obj: T, start_fn: ResourceStartFn, stop_fn: ResourceStopFn,
        activity_fn: ResourceActivityFn, usage_fn: ResourceUsageFn,
    ):
        self._obj = obj
        self._start_fn = start_fn
        self._stop_fn = stop_fn
        self._activity_fn = activity_fn
        self._usage_fn = usage_fn

    @property
    def active(self) -> bool:
        return self._activity_fn(self._obj)

    def usages(self) -> int:
        return self._usage_fn(self._obj)

    def start(self) -> Awaitable[None] | None:
        return self._start_fn(self._obj)

    def stop(self) -> Awaitable[None] | None:
        return self._stop_fn(self._obj)


def make_resource[T](
    obj: T, start_fn: ResourceStartFn, stop_fn: ResourceStopFn,
    activity_fn: ResourceActivityFn, usage_fn: ResourceUsageFn,
) -> ResourceAdapter[T]:
    return ResourceAdapter[T](obj, start_fn, stop_fn, activity_fn, usage_fn)



from typing import Any, Pattern

from pydantic import BaseModel, ConfigDict, Field

from ..entities.operator import OperatorKey
from ..entities.providers import ProviderKey


class DepsPolicy[DK](BaseModel):
    model_config = ConfigDict(frozen=True)

    key: DK
    name_pattern: Pattern | None = None


class ResourcePolicy(DepsPolicy[Any]):
    pass


class ProviderPolicy(DepsPolicy[ProviderKey]):
    pass


class OperatorPolicy(DepsPolicy[OperatorKey]):
    pass


class ResourceConf[K](BaseModel):
    model_config = ConfigDict(frozen=True)

    key: K
    name: str | None = None
    args: dict[str, Any] = Field(default_factory=dict)
    enabled: bool = Field(default=True)
    resources_read: list[ResourcePolicy] = Field(default_factory=list)
    providers_read: list[ProviderPolicy] = Field(default_factory=list)
    operators_read: list[OperatorPolicy] = Field(default_factory=list)



from typing import Any, Callable, Optional, Pattern

from pydantic import BaseModel, ConfigDict, Field, PrivateAttr

from greyhorse.utils.invoke import caller_path
from .controller import ControllerConf
from .service import ServiceConf
from ..entities.controller import ControllerFactoryMapping
from ..entities.module import Module
from ..entities.operator import OperatorKey
from ..entities.providers import ProviderKey
from ..entities.service import ServiceFactoryMapping


class ProviderClaim(BaseModel):
    model_config = ConfigDict(frozen=True)

    key: ProviderKey
    name_pattern: Pattern | None = None


class OperatorExport(BaseModel):
    model_config = ConfigDict(frozen=True)

    key: OperatorKey
    name_pattern: Pattern | None = None


class ModuleDesc(BaseModel):
    path: str = Field(frozen=True)
    args: dict[str, Any] = Field(default_factory=dict)
    enabled: bool = Field(default=True, frozen=True)

    _conf: Optional['ModuleConf'] = PrivateAttr(default=None)
    _initpath: list[str] = PrivateAttr(default_factory=lambda: caller_path(5))


class ModuleConf(BaseModel):
    model_config = ConfigDict(frozen=True)

    name: str = Field(frozen=True)
    enabled: bool = Field(default=True)
    factory: Callable[[...], Module] = Field(default=Module, frozen=True)

    submodules: list[ModuleDesc] = Field(default_factory=list)

    controllers: list[ControllerConf] = Field(default_factory=list)
    services: list[ServiceConf] = Field(default_factory=list)

    controller_factories: ControllerFactoryMapping = Field(default_factory=dict)
    service_factories: ServiceFactoryMapping = Field(default_factory=dict)

    provider_claims: list[ProviderClaim] = Field(default_factory=list)
    operator_exports: list[OperatorExport] = Field(default_factory=list)



from abc import ABC, abstractmethod
from typing import Any, Callable, Generic, Mapping, Sequence, Tuple, Type, TypeVar

from ..models.model import Model
from ...utils.invoke import is_awaitable

IdType = TypeVar('IdType')
EntityType = TypeVar('EntityType')
ModelType = TypeVar('ModelType', bound=Model, covariant=True)
EntityFactory = Callable[..., EntityType]
ModelFactory = Callable[..., ModelType]


class Repository(Generic[IdType, EntityType], ABC):
    def __init__(self, class_: Type[EntityType], factory: EntityFactory | None = None):
        self._class = class_
        self._factory = factory or class_

    @property
    def entity_class(self):
        return self._class

    @property
    def entity_factory(self):
        return self._factory

    async def construct(self, data: Mapping[str, Any], **kwargs) -> EntityType:
        result = self.entity_factory(**data)
        if is_awaitable(result):
            return await result
        return result

    @abstractmethod
    async def get(self, id_value: IdType, **kwargs) -> EntityType | None:
        ...

    @abstractmethod
    async def get_any(self, indices: Sequence[IdType], **kwargs) -> Sequence[EntityType | None]:
        ...

    @abstractmethod
    async def exists(self, id_value: IdType, **kwargs) -> bool:
        ...

    @abstractmethod
    async def load(self, instance: EntityType, only: Sequence[str] | None = None) -> bool:
        ...

    @abstractmethod
    async def create(self, data: Mapping[str, Any], **kwargs) -> EntityType | None:
        ...

    async def get_or_create(
        self, id_value: IdType, data: Mapping[str, Any], **kwargs,
    ) -> Tuple[EntityType | None, bool]:
        if instance := await self.get(id_value, **kwargs):
            return instance, False
        else:
            return await self.create(data, **kwargs), True

    @abstractmethod
    async def update_by_id(self, id_value: IdType, data: Mapping[str, Any], **kwargs) -> bool:
        ...

    @abstractmethod
    async def save(self, instance: EntityType, **kwargs) -> bool:
        ...

    @abstractmethod
    async def save_all(self, objects: Sequence[EntityType], **kwargs) -> bool:
        ...

    @abstractmethod
    async def delete(self, instance: EntityType) -> bool:
        ...

    @abstractmethod
    async def delete_all(self, indices: Sequence[IdType] | None = None) -> int:
        ...

    @abstractmethod
    async def delete_by_id(self, id_value: IdType) -> bool:
        ...


class ModelRepository(Repository[IdType, ModelType], ABC):
    def __init__(
        self, model_class: Type[ModelType],
        model_factory: ModelFactory | None = None,
    ):
        super().__init__(model_class, model_factory)
        model_class.bind(self)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import timedelta
from typing import Any, AsyncGenerator, Mapping, Sequence, Tuple


@dataclass
class CacheData:
    cache_id: str
    data: Any | None = None


class CacheOperator(ABC):
    def get_cache_id(self, id_value: Any) -> str:
        return str(id_value)

    @abstractmethod
    def get_cache_key(self, id_value: Any) -> str:
        ...

    @abstractmethod
    async def cache_one(self, data: CacheData, ttl: timedelta | None = None, **kwargs) -> bool:
        ...

    @abstractmethod
    async def cache_many(self, objects: Sequence[CacheData], ttl: timedelta | None = None, **kwargs) -> bool:
        ...

    @abstractmethod
    async def load_one(self, cache_key: str, **kwargs) -> Tuple[bool, Any | None]:
        ...

    @abstractmethod
    async def load_many(self, indices: Sequence[Any], **kwargs) -> Sequence[Mapping[str, Any] | None]:
        ...

    @abstractmethod
    async def drop_one(self, cache_key: str) -> bool:
        ...

    @abstractmethod
    async def drop_many(self, indices: Sequence[Any] | None = None) -> int:
        ...

    @abstractmethod
    async def drop_all(self, prefix: str | None = None) -> int:
        ...

    @abstractmethod
    async def exists(self, cache_key: str) -> bool:
        ...


class ModelCacheOperator(CacheOperator):
    @abstractmethod
    def get_model_cache_key(self) -> str:
        ...

    def get_cache_id(self, id_value: Any) -> str:
        return str(id_value)

    def get_cache_key(self, id_value: Any) -> str:
        return f'{self.get_model_cache_key()}:{self.get_cache_id(id_value)}'


class ScanCacheMixin(CacheOperator):
    @abstractmethod
    async def scan(
        self, cache_key: str | None = None, match: str | None = None,
            count: int | None = None) -> AsyncGenerator[Tuple[str, Any | None], None]:
        ...



from functools import partial
from importlib import import_module


def import_path(dotted_path: str):
    if ':' in dotted_path:
        module_path, attr_path = dotted_path.rsplit(':', maxsplit=1)
    else:
        module_path, attr_path = dotted_path.rsplit('.', maxsplit=1)

    module = import_module(module_path)

    if ':' in dotted_path:
        result = module
        for attr_name in attr_path.rsplit('.'):
            result = getattr(result, attr_name)
        return result

    return getattr(module, attr_path)


def lazy_import(dotted_path: str, callable: bool = False, as_partial: bool = False):
    if callable:
        def inner(*args, **kwargs):
            if as_partial:
                return partial(import_path(dotted_path), *args, **kwargs)
            return import_path(dotted_path)(*args, **kwargs)
    else:
        def inner():
            return import_path(dotted_path)
    return inner



from abc import abstractmethod
from typing import Any, Generic, Mapping, Sequence, TypeVar

from .base import EntityType, IdType, Repository

FilterType = TypeVar('FilterType')
SortingType = TypeVar('SortingType')


class FilterableRepository(
    Repository[IdType, EntityType],
    Generic[IdType, EntityType, FilterType, SortingType],
):
    @abstractmethod
    async def list(
        self, filters: FilterType | None = None,
        sorting: SortingType | None = None,
        skip: int = 0, limit: int = 0, **kwargs,
    ) -> Sequence[EntityType]:
        ...

    @abstractmethod
    async def sublist(
        self, field, filters: FilterType | None = None,
        sorting: SortingType | None = None,
        skip: int = 0, limit: int = 0, **kwargs,
    ) -> Sequence[EntityType]:
        ...

    @abstractmethod
    async def count(self, filters: FilterType | None = None, **kwargs) -> int:
        ...

    @abstractmethod
    async def exists_by(self, filters: FilterType, **kwargs) -> bool:
        ...

    @abstractmethod
    async def update_by(self, filters: FilterType, data: Mapping[str, Any], **kwargs) -> int:
        ...

    @abstractmethod
    async def delete_by(self, filters: FilterType, **kwargs) -> int:
        ...

    #
    # Query operations
    #

    @abstractmethod
    def query_for_select(self, **kwargs):
        ...

    @abstractmethod
    def query_for_update(self, **kwargs):
        ...

    @abstractmethod
    def query_for_delete(self, **kwargs):
        ...

    @abstractmethod
    def query_get(self, id_value: IdType, query=None, **kwargs):
        ...

    @abstractmethod
    def query_any(self, indices: Sequence[IdType], query=None, **kwargs):
        ...

    @abstractmethod
    def query_list(
        self, filters: FilterType | None = None,
        sorting: SortingType | None = None,
        skip: int = 0, limit: int = 0, **kwargs,
    ):
        ...

    @abstractmethod
    def query_count(self, filters: FilterType | None = None, **kwargs):
        ...

    @abstractmethod
    def query_exists(self, id_value: IdType, **kwargs):
        ...

    @abstractmethod
    def query_exists_by(self, filters: FilterType, **kwargs):
        ...

    @abstractmethod
    def query_update(self, filters: FilterType, **kwargs):
        ...

    @abstractmethod
    def query_delete(self, filters: FilterType, **kwargs):
        ...



from datetime import timedelta
from math import floor

from pytimeparse.timeparse import timeparse


def parse_timeout(value: str, default: str = '30 secs') -> timedelta:
    return timedelta(seconds=timeparse(value or default))


def calc_next_interval(
    initial_period: timedelta, iteration: int,
    iteration_divisor: int = 1, max_seconds: float = 3600,
) -> float:
    """
    The function counts the next time interval in seconds with progressive increase

    initial_period: Initial value of the period.
    iteration: Current iteration number.
    iteration_divisor: Iteration divider, for the ability to change the interval not every iteration, but after several.
    max_seconds: Maximum period duration in seconds.
    """
    def calc_interval(initial: float, iterations: int):
        f = initial
        for i in range(iterations):
            f *= 2.0 if i % 2 == 0 else 1.5
        return f

    initial_period = initial_period.total_seconds()
    next_interval = min(calc_interval(initial_period, floor(iteration / iteration_divisor)), max_seconds)
    return next_interval



import asyncio
import threading
from abc import ABC, abstractmethod
from typing import Awaitable, Callable, Mapping, TYPE_CHECKING, cast

from greyhorse.result import Result
from .deps import DepsProvider
from .operator import OperatorFactoryFn, OperatorFactoryRegistry, OperatorKey
from .providers import ProviderFactoryFn, ProviderFactoryRegistry, ProviderKey
from ..errors import NoOpFoundForPattern, NoProvFoundForPattern, OpPolicyViolation, ProvPolicyViolation
from ..utils.registry import DictRegistry, KeyMapping
from ...i18n import tr
from ...logging import logger

if TYPE_CHECKING:
    from ..schemas.service import ProviderMappingPolicy
    from ..schemas.controller import OperatorMappingPolicy

type ServiceKey = type[Service]
type ServiceFactoryFn = Callable[[...], Result[Service]]
type ServiceFactoryMapping = Mapping[ServiceKey, ServiceFactoryFn]


class Service(ABC):
    def __init__(
        self, name: str, deps_provider: DepsProvider | None = None,
    ):
        self._name = name
        self._deps_provider = deps_provider
        self._op_factories = DictRegistry[OperatorKey, OperatorFactoryFn]()
        self._provider_factories = DictRegistry[ProviderKey, ProviderFactoryFn]()

    @property
    def name(self) -> str:
        return self._name

    @property
    @abstractmethod
    def active(self) -> bool:
        ...

    def set_deps_provider(self, instance: DepsProvider):
        self._deps_provider = instance

    def reset_deps_provider(self):
        self._deps_provider = None

    def get_operator_factory(self, key: OperatorKey, name: str | None = None) -> OperatorFactoryFn | None:
        if not self._deps_provider:
            return None
        return self._deps_provider.get_operator_factory(key, name=name)

    def get_provider_factory(self, key: ProviderKey, name: str | None = None) -> ProviderFactoryFn | None:
        if not self._deps_provider:
            return None
        return self._deps_provider.get_provider_factory(key, name=name)

    def get_resource[T](self, key: type, name: str | None = None) -> T | None:
        if not self._deps_provider:
            return None
        if instance := self._deps_provider.get_resource(key, name=name):
            return cast(T, instance)
        return None

    @property
    def provider_factories(self) -> ProviderFactoryRegistry:
        return self._provider_factories

    @property
    def operator_factories(self) -> OperatorFactoryRegistry:
        return self._op_factories

    def create(self) -> Awaitable[Result] | Result:
        return Result.from_ok()

    def destroy(self) -> Awaitable[Result] | Result:
        return Result.from_ok()

    def start(self) -> Awaitable[None] | None:
        pass

    def stop(self) -> Awaitable[None] | None:
        pass

    @abstractmethod
    def wait(self) -> threading.Event | asyncio.Event:
        ...

    def check_operator_mapping(
        self, policies: list['OperatorMappingPolicy'],
    ) -> Result[Mapping[OperatorKey, KeyMapping[OperatorKey]]]:
        policies_dict = {p.key: p for p in policies}
        key_mapping = {}

        for key in self._op_factories.list_keys():
            if key not in policies_dict:
                error = OpPolicyViolation(type='service', name=self.name, key=str(key.__name__))
                logger.error(error.message)
                return Result.from_error(error)

            name_pattern = policies_dict[key].name_pattern

            if name_pattern is None:
                key_mapping[key] = KeyMapping[OperatorKey](map_to=policies_dict[key].map_to)
                policies_dict.pop(key)
                continue

            names = []

            for name in self._op_factories.get_names(key):
                if name_pattern.match(name):
                    names.append(name)

            if not names:
                error = NoOpFoundForPattern(
                    type='service', name=self.name, key=str(key.__name__), pattern=name_pattern,
                )
                logger.error(error.message)
                return Result.from_error(error)

            key_mapping[key] = KeyMapping[OperatorKey](map_to=policies_dict[key].map_to, names=names)
            policies_dict.pop(key)

        for p in policies_dict.values():
            logger.warn(tr('app.entities.operator-unused-policy').format(
                type='service', name=self.name, key=str(p.key.__name__), pattern=p.name_pattern or '-',
            ))

        return Result.from_ok(key_mapping)

    def check_provider_mapping(
        self, policies: list['ProviderMappingPolicy'],
    ) -> Result[Mapping[ProviderKey, KeyMapping[ProviderKey]]]:
        policies_dict = {p.key: p for p in policies}
        key_mapping = {}

        for key in self._provider_factories.list_keys():
            if key not in policies_dict:
                error = ProvPolicyViolation(type='service', name=self.name, key=str(key.__name__))
                logger.error(error.message)
                return Result.from_error(error)

            name_pattern = policies_dict[key].name_pattern

            if name_pattern is None:
                key_mapping[key] = KeyMapping[ProviderKey](map_to=key)
                policies_dict.pop(key)
                continue

            names = []

            for name in self._provider_factories.get_names(key):
                if name_pattern.match(name):
                    names.append(name)

            if not names:
                error = NoProvFoundForPattern(
                    type='service', name=self.name, key=str(key.__name__), pattern=name_pattern,
                )
                logger.error(error.message)
                return Result.from_error(error)

            key_mapping[key] = KeyMapping[ProviderKey](map_to=key, names=names)
            policies_dict.pop(key)

        for p in policies_dict.values():
            logger.warn(tr('app.entities.provider-unused-policy').format(
                type='service', name=self.name, key=str(p.key.__name__), pattern=p.name_pattern or '-',
            ))

        return Result.from_ok(key_mapping)



import enum
from abc import ABC, abstractmethod
from contextlib import AbstractAsyncContextManager, AbstractContextManager, asynccontextmanager, contextmanager
from typing import Callable

from greyhorse.result import Result
from ..utils.registry import ReadonlyRegistry

type ProviderKey = type[Provider]
type ProviderFactoryFn = Callable[[...], Result[Provider]]
type ProviderFactoryRegistry = ReadonlyRegistry[ProviderKey, ProviderFactoryFn]


class BorrowType(str, enum.Enum):
    SIMPLE = 'simple'
    SESSION = 'session'
    EXCLUSIVE = 'exclusive'
    SHARED = 'shared'


class Provider:
    is_sync: bool
    is_async: bool
    type: BorrowType

    @property
    def active(self) -> bool:
        return False


class SyncSimpleProvider[R](Provider, ABC):
    is_sync = True
    is_async = False
    type = BorrowType.SIMPLE

    @abstractmethod
    def get(self, *args, **kwargs) -> R:
        ...


class AsyncSimpleProvider[R](Provider, ABC):
    is_sync = False
    is_async = True
    type = BorrowType.SIMPLE

    @abstractmethod
    async def get(self, *args, **kwargs) -> R:
        ...


class SyncSessionProvider[R](Provider, ABC):
    is_sync = True
    is_async = False
    type = BorrowType.SESSION

    @contextmanager
    @abstractmethod
    def acquire_session(self, *args, **kwargs) -> AbstractContextManager[R]:
        ...

    def release(self):
        pass


class AsyncSessionProvider[R](Provider, ABC):
    is_sync = False
    is_async = True
    type = BorrowType.SESSION

    @asynccontextmanager
    @abstractmethod
    async def acquire_session(self, *args, **kwargs) -> AbstractAsyncContextManager[R]:
        ...

    async def release(self):
        pass


class SyncExclusiveProvider[R](Provider, ABC):
    is_sync = True
    is_async = False
    type = BorrowType.EXCLUSIVE

    @abstractmethod
    def acquire(self, *args, **kwargs) -> Result[R]:
        ...

    @abstractmethod
    def release(self, instance: R) -> Result:
        ...


class AsyncExclusiveProvider[R](Provider, ABC):
    is_sync = False
    is_async = True
    type = BorrowType.EXCLUSIVE

    @abstractmethod
    async def acquire(self, *args, **kwargs) -> Result[R]:
        ...

    @abstractmethod
    async def release(self, instance: R) -> Result:
        ...


class SyncSharedProvider[ExclR, SharedR](Provider, ABC):
    is_sync = True
    is_async = False
    type = BorrowType.SHARED

    @abstractmethod
    def acquire_exclusive(self, *args, **kwargs) -> Result[ExclR]:
        ...

    @abstractmethod
    def acquire_shared(self, *args, **kwargs) -> Result[SharedR]:
        ...

    @abstractmethod
    def release(self, instance: ExclR | SharedR) -> Result:
        ...


class AsyncSharedProvider[ExclR, SharedR](Provider, ABC):
    is_sync = False
    is_async = True
    type = BorrowType.SHARED

    @abstractmethod
    async def acquire_exclusive(self, *args, **kwargs) -> Result[ExclR]:
        ...

    @abstractmethod
    async def acquire_shared(self, *args, **kwargs) -> Result[SharedR]:
        ...

    @abstractmethod
    async def release(self, instance: ExclR | SharedR) -> Result:
        ...



from abc import ABC, abstractmethod
from typing import Any

from .operator import OperatorFactoryFn, OperatorKey
from .providers import ProviderFactoryFn, ProviderKey


class DepsProvider(ABC):
    @abstractmethod
    def get_operator_factory(
        self, key: OperatorKey, name: str | None = None,
    ) -> OperatorFactoryFn | None:
        ...

    @abstractmethod
    def get_provider_factory(
        self, key: ProviderKey, name: str | None = None,
    ) -> ProviderFactoryFn | None:
        ...

    @abstractmethod
    def get_resource(
        self, key: Any, name: str | None = None,
    ) -> Any | None:
        ...


class DepsOperator(ABC):

    @abstractmethod
    def set_resource(
        self, key: Any, instance: Any, name: str | None = None,
    ) -> bool:
        ...



import inspect
from typing import Any

from greyhorse.result import Result
from greyhorse.utils.injectors import ParamsInjector

from ..utils.registry import ReadonlyRegistry
from ..entities.providers import ProviderFactoryRegistry, ProviderKey
from ..entities.operator import OperatorFactoryRegistry, OperatorKey
from ..errors import ModuleFunctionError
from ..schemas.components import DepsPolicy, ResourceConf


class ResourceBuilder[K, R]:
    def __init__(
        self,
        deps_registry: ReadonlyRegistry[Any, Any],
        providers_registry: ProviderFactoryRegistry,
        operators_registry: OperatorFactoryRegistry,
        mapping  #: ResFactoryMapping[K, R],
    ):
        self._mapping = mapping
        self._deps_registry = deps_registry
        self._provider_factories = providers_registry
        self._operator_factories = operators_registry

    def _filter_deps(self, deps_filters: list[DepsPolicy], registry: ReadonlyRegistry) -> dict:
        result = {}

        for deps_filter in deps_filters:
            if deps_filter.name_pattern is not None:
                for name in registry.get_names(deps_filter.key):
                    if deps_filter.name_pattern.match(name):
                        result[deps_filter.key] = registry.get(deps_filter.key, name)
                        break
            elif dep := registry.get(deps_filter.key):
                result[deps_filter.key] = dep

            if deps_filter.key not in result:
                pass  # TODO error operator should provide "deps_filter.key"

        return result

    def __call__(self, conf: ResourceConf[K], **kwargs) -> Result[R]:
        if not conf.enabled:
            return Result.from_ok()  # TODO

        injector = ParamsInjector()

        values = {'name': conf.name, **conf.args}
        values.update(kwargs)

        for key, value in self._filter_deps(conf.deps, self._deps_registry).items():
            if inspect.isclass(key):
                injector.add_type_provider(key, value)
            else:
                values[key] = value

        for deps_conf, registry in (
            (conf.providers_deps, self._provider_factories),
            (conf.operators_deps, self._operator_factories),
        ):
            for key, factory in self._filter_deps(deps_conf, registry).items():
                res = factory()
                if not res.success:
                    return res
                injector.add_type_provider(key, res.result)

        if factory := self._mapping.get(conf.key):
            injected_args = injector(factory, values=values)
            res = factory(*injected_args.args, **injected_args.kwargs)
            if not res.success:
                return res
            instance = res.result
        else:
            return Result.from_error(None)  # TODO

        return Result.from_ok(instance)



import dataclasses
import enum
from functools import wraps
from typing import Callable, Container, Optional, Protocol, Self, Sequence, Type

from .i18n import tr


class ResultProtocol[T](Protocol):
    success: bool
    result: T | None
    errors: Sequence['Error'] | None


class ResultStatus(str, enum.Enum):
    OK: str = 'ok'
    PARTIAL: str = 'partial'
    ERROR: str = 'error'


@dataclasses.dataclass(slots=True, frozen=True)
class Result[T]:
    success: bool
    result: T | None = None
    errors: Sequence['Error'] | None = dataclasses.field(default_factory=list)

    @classmethod
    def from_ok(cls, value: T | None = None):
        return cls(success=True, result=value, errors=[])

    @classmethod
    def from_error(cls, error: 'Error'):
        return cls(success=False, result=None, errors=[error])

    @classmethod
    def from_errors(cls, errors: Sequence['Error']):
        return cls(success=False, result=None, errors=errors)

    @property
    def error(self) -> Optional['Error']:
        return self.errors[0] if len(self.errors) else None

    @property
    def error_count(self) -> int:
        return len(self.errors)


@dataclasses.dataclass(slots=True, frozen=True)
class ComplexResult[T: Container]:
    status: ResultStatus
    result: T | None = None
    errors: Sequence['Error'] | None = dataclasses.field(default_factory=list)

    @classmethod
    def from_ok(cls, values: T | None = None):
        return cls(status=ResultStatus.OK, result=values, errors=[])

    @classmethod
    def from_partial(
        cls, values: T | None = None, error: Optional['Error'] = None,
        errors: Sequence['Error'] | None = None,
    ):
        errors = errors if errors else [error] if error else []
        return cls(status=ResultStatus.PARTIAL, result=values, errors=errors)

    @classmethod
    def from_error(cls, error: 'Error'):
        return cls(status=ResultStatus.ERROR, result=None, errors=[error])

    @classmethod
    def from_errors(cls, errors: Sequence['Error']):
        return cls(status=ResultStatus.ERROR, result=None, errors=errors)

    @property
    def error(self) -> Optional['Error']:
        return self.errors[0] if len(self.errors) else None

    @property
    def error_count(self) -> int:
        return len(self.errors)


type ListResult[T] = ComplexResult[list[T]]
type DictResult[K, V] = ComplexResult[dict[K, V]]


class Error:
    _type_classes: dict[str, type[Self]] = dict()

    app: str = ''
    type: str
    msg: str = ''
    tr_key: str = ''

    @property
    def message(self):
        if self.tr_key:
            return tr('.'.join([self.app, self.tr_key]))
        return tr('.'.join([self.app, self.type]), default=self.msg)

    def __repr__(self):
        return f'Error [{self.app}] ({self.type}): \"{self.message}\"'

    def __eq__(self, other: Self):
        return (self.app, self.type) == (other.app, other.type)

    def __hash__(self):
        return hash((self.app, self.type))

    @property
    def dict(self):
        return dict(app=self.app, type=self.type, message=self.message)

    @classmethod
    def get_by_type(cls, type_: str) -> Type[Self] | None:
        return cls._type_classes.get(type_)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if hasattr(cls, 'type'):
            cls._type_classes[cls.type] = cls


class ErrorKwargsMixin:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    # noinspection PyUnresolvedReferences
    @property
    def message(self):
        message = super(ErrorKwargsMixin, self).message
        return message.format(**self.kwargs)


def result_or(result, error):
    if callable(result):
        result = result()
    if result:
        return result
    if isinstance(error, Exception):
        if callable(error):
            raise error()
        raise error
    return error


# noinspection PyPep8Naming
def F[T, **P](func: Callable[P, T]) -> Callable[P, Result[T]]:
    @wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs) -> Result[T]:
        return Result.from_ok(func(*args, **kwargs))
    return inner



from abc import ABC, abstractmethod
from typing import Mapping

from .base import AsyncEngine, SyncEngine


class EngineFactory(ABC):
    @abstractmethod
    def get_engine_names(self) -> list[str]:
        ...


class SyncEngineFactory(EngineFactory):
    @abstractmethod
    def __call__(self, name: str, *args, **kwargs) -> SyncEngine:
        ...

    @abstractmethod
    def get_engine(self, name: str) -> SyncEngine | None:
        ...

    @abstractmethod
    def get_engines(self, names: list[str] | None = None) -> Mapping[str, SyncEngine]:
        ...


class AsyncEngineFactory(EngineFactory):
    @abstractmethod
    def __call__(self, name: str, *args, **kwargs) -> AsyncEngine:
        ...

    @abstractmethod
    def get_engine(self, name: str) -> AsyncEngine | None:
        ...

    @abstractmethod
    def get_engines(self, names: list[str] | None = None) -> Mapping[str, AsyncEngine]:
        ...



import asyncio
import operator
from datetime import timedelta
from functools import reduce
from typing import Any, Mapping, Sequence, Tuple

from greyhorse.data.cache.base import CacheData, ModelCacheOperator
from greyhorse.utils.hashes import calculate_digest


class MethodCache:
    def __init__(self, op: ModelCacheOperator, ttl: timedelta | None = None):
        self._op = op
        self._ttl = ttl

    @property
    def op(self) -> ModelCacheOperator:
        return self._op

    def get_model_cache_key(self) -> str:
        return self._op.get_model_cache_key()

    def get_cache_id(self, id_value: Any) -> str:
        return self._op.get_cache_id(id_value)

    def get_cache_key(self, id_value: Any) -> str:
        return self._op.get_cache_key(id_value)

    @staticmethod
    def cache_key_for(method: str, args: Mapping[str, Any] | None = None) -> str:
        id_data = [method]

        if args:
            id_data.append(calculate_digest(args))
        else:
            id_data.append('_')

        return ':'.join(id_data)

    async def put(self, method: str, args: Mapping[str, Any] | None = None,
                  data: Any | None = None, cache_key: str | None = None,
                  ttl: timedelta | None = None, **kwargs) -> bool:
        cache_key = cache_key or self.cache_key_for(method, args)
        data = CacheData(cache_id=cache_key, data=data)
        return await self._op.cache_one(data, ttl=ttl or self._ttl, **kwargs)

    async def get(self, method: str, args: Mapping[str, Any] | None = None,
                  cache_key: str | None = None, **kwargs) -> Tuple[bool, Any | None]:
        cache_key = cache_key or self.cache_key_for(method, args)
        exists, data = await self._op.load_one(self.get_cache_key(cache_key), **kwargs)
        return exists, data

    async def drop(self, method: str, args: Mapping[str, Any] | None = None,
                   cache_key: str | None = None) -> bool:
        cache_key = cache_key or self.cache_key_for(method, args)
        return await self._op.drop_one(self.get_cache_key(cache_key))

    async def drop_all(self, methods: Sequence[str] | None = None) -> int:
        methods = methods if methods else list()
        if not methods:
            return await self._op.drop_all()
        results = await asyncio.gather(
            *(self._op.drop_all(method) for method in methods)
        )
        return reduce(operator.add, results, 0)
